--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

require 'middleclass'

Vector2d = class('Vector2d')

-- Creates a new Vector2d whose origin is assumed to be 0, 0.
-- @param x Horizontal component of the vector.
-- @param y Vertical component of the vector.
-- @return Newly created Vector2d.
function Vector2d:initialize(x, y)
	self.x = x 
  self.y = y
end

-- Converts a Vector2d to a string representation.
-- @param self The Vector2d.
-- @return String representation of the Vector2d.
function Vector2d:tostring()
	return "("..tonumber(self.x)..","..tonumber(self.y)..")"
end

-- Scales the Vector2d by the given scalar value.
-- @param self The Vector2d.
-- @param scale Scalar value to scale the vector by.
-- @return A new Vector2d scaled by the given value.
function Vector2d:scale(scale)
	assert(self and scale, "missing argument")
	return Vector2d(scale * self.x, scale * self.y)
end

-- Divides a Vector2d by the given scalar value.
-- @param self The Vector2d
-- @param scale Scalar value to divide the vector by.
-- @param A new Vector2d divided by the given value.
function Vector2d:divide(scale)
	assert(self and scale, "missing argument")
	return Vector2d(self.x / scale, self.y / scale)
end

-- Adds one vector to another.
-- @param self The Vector2d.
-- @param vector Vector2d to add.
-- @return New Vector2d that is the result of the addition.
function Vector2d:add(vector)
	assert(self and vector, "missing argument")
	return Vector2d(self.x + vector.x, self.y + vector.y)
end

-- Subtracts one vector from another.
-- @param self The Vector2d.
-- @param vector Vector2d to subtract.
-- @return New Vector2d that is the result of the subtraction.
function Vector2d:subtract(vector)
	assert(self and vector, "missing argument")
	return Vector2d(self.x - vector.x, self.y - vector.y)
end

-- Computes the coodinatewise product of the two vectors.
-- @param self The Vector2d.
-- @param vector Vector2d to multiply with self.
-- @return New Vector2d that is the result of oordinatewise multiplication.
function Vector2d:coordinateProduct(vector)
	return Vector2d(self.x * vector.x, self.y * vector.y)
end

-- Computes the dot product of the two vectors.
-- @param self The Vector2d.
-- @param vector Vector2d to dot with self.
-- @return Dot product.
function Vector2d:dotProduct(vector)
	return self.x * vector.x + self.y * vector.y
end

-- Computes the determinant of this Vector2d and the parameter Vector2d.
-- @param self The Vector2d.
-- @param vector Other Vector2d.
-- @return Determinant of the two vectors.
function Vector2d:determinant(vector)
	return self.x * vector.y - self.y * vector.x
end

-- Indicates whether the Vector2d is equal to the parameter vector.
-- @param self The Vector2d.
-- @param vector Vector2d to compare.
-- @return true if the Vector2d is equal to the parameter vector.
function Vector2d:eq(vector)
	return self.x == vector.x and self.y == vector.y
end

-- Indicates whether the Vector2d is less than the parameter vector.
-- @param self The Vector2d.
-- @param vector Vector2d to compare.
-- @return true if the Vector2d is less than the parameter vector.
function Vector2d:lt(vector)
	return self.x < vector.x or (self.x == vector.x and self.y < vector.y)
end

-- Indicates whether the Vector2d is less than or equal to the parameter vector.
-- @param self The Vector2d.
-- @param vector Vector2d to compare.
-- @return true if the Vector2d is less than or equal to the parameter vector.
function Vector2d:le(vector)
	return self.x <= vector.x and self.y <= vector.y
end

-- Gets the squared length of the Vector2d.
-- @param self The Vector2d.
-- @return Length of the Vector2d squared.
function Vector2d:length2()
	return self.x * self.x + self.y * self.y
end

-- Gets the length of the Vector2d.
-- @param self The Vector2d.
-- @return Length of the Vector2d.
function Vector2d:length()
	return math.sqrt(self.x * self.x + self.y * self.y)
end

-- Gets the distance between the points at the end of each Vector2d.
-- @param self The Vector2d.
-- @param vector Second Vector2d.
-- @return Distance between the points of the two vectors.
function Vector2d:distance(vector)
	return length(Vector2d(self.x - vector.x, self.y - vector.y))
end

-- Returns a normalized version of the given Vector2d.
-- @param self The Vector2d.
-- @return Normalized Vector2d.
function Vector2d:normalize()
	local l = self:length()
	if l > 0 then
		return Vector2d(self.x/l, self.y/l)
	end
	return Vector2d(self.x, self.y)
end

-- Rotates a vector by the given angle.
-- @param self The Vector2d.
-- @param phi Angle in radians.
-- @return Resulting rotated Vector2d.
function Vector2d:rotate(phi)
	local c, s = math.cos(phi), math.sin(phi)
	return Vector2d(c* self.x - s * self.y, s * self.x + c * self.y)
end

-- Returns a Vector2d perpindicular to the original.
-- @param self The Vector2d.
-- @return Perpindicular vector.
function Vector2d:perpendicular()
	return Vector2d(-self.y, self.x)
end

-- Projects the Vector2d onto the given vector and returns the result.
-- @param self The Vector2d.
-- @param vector Vector2d to be projected.
-- @return Resulting Vector2d.
function Vector2d:project(vector)
	local s = (self.x * vector.x + self.y * vector.y) / (vector.x * vector.x + vector.y * vector.y)
	return Vector2d(s * vector.x, s * vector.y)
end

-- Mirrors the Vector2d along the given vector and returns the result.
-- @param self The Vector2d.
-- @param vector Used as the mirror axis.
-- @return New mirrored Vector2d.
function Vector2d:mirror(vector)
	local s = 2 * (self.x * vector.x + self.y * vector.y) / (vector.x * vector.x + vector.y * vector.y)
	return Vector2d(s * vector.x - self.x, s * vector.y - self.y)
end

-- Gets the angle of the Vector2d in radians.
-- @param self The Vector2d
-- @returns Angle of the vector.
function Vector2d:angle()
	assert(self, 'missing paramter [self]')
	return math.atan2(self.y, self.x)
end

-- Gets the angle facing the opposite direction of the Vector2d in radians.
-- @param self The Vector2d
-- @returns Opposing angle of the vector.
function Vector2d:opposingAngle()
	assert(self, 'missing parameter [self]')
	return math.atan2(self.y, self.x) + math.pi
end

-- Copies an existing vector instance.
-- @param self The Vector2d.
-- @return Deep copy of the vector.
function Vector2d:copy()
	return Vector2d(self.x, self.y)
end

